
conda env list

conda create --name enbios37test --clone enbios37

spolddir =  'C:\Users\1150226\Universitat Autònoma de Barcelona\Sentinel Sostenipra - SENTINEL\SHARED\WP2 -  Socio Environmental\Data\Ecospold\APOS LCI'

pip install enbios==0.7